import sys
import logging
from collections import defaultdict
from pathlib import Path

import esmvalcore.experimental as esmvaltool
from esmvalcore.dataset import Dataset
from esmvalcore.local import LocalFile


# TODO allow setting different config files in the same way that
# ESMValTool does using an argument parser

# TODO setting logger default level from command line
logging.captureWarnings(True)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def copy2gpfs(recipe_file, tfolder="/gpfs/scratch/bsc32/$USER"):
    """
    Generate instructions to copy ESMValTool recipe required files from
    esarchive to GPFS.

    Parameters
    ----------
    recipe_file: str or pathlib.Path
      Path to the ESMValTool recipe file.
    tfolder: str or pathlib.Path (optional)
      Folder to save the data in. Could be in 'scratch' or 'projects'.
      This will be saved in an environment variable in the top of the
      file and can be easily changed dending on the occupation. Default
      is '/gpfs/scratch/bsc32/$USER' which will point to your user folder
      in /gpfs/scratch.

    """
    recipe = esmvaltool.get_recipe(recipe_file)
    datasets = recipe.data.get('datasets', [])

    # Save required files in sets to avoid 'bringing' them twice
    required_files = defaultdict(set)
    # Loop over diagnostics
    for diag in recipe.data['diagnostics'].values():
        if 'variables' not in diag or diag['variables'] is None:
            # Some diagnostics may not have variables defined,
            # in that case move to the next
            continue
        # Loop over variables in a diagnostic
        for var, var_values in diag['variables'].items():
            # TODO build also auxiliary vars
            if 'short_name' not in var_values:
                # Add short_name if not given
                var_values['short_name'] = var
            # Define the target datasets (main + additional)
            var_datasets = datasets + var_values.get('additional_datasets', [])
            # Loop over target datasets
            for ds_values in var_datasets:
                # Define the dataset object
                ds_obj = Dataset(**var_values, **ds_values)
                # find the files
                ds_obj.find_files()
                # Get only LocalFiles resolving their paths
                files = [
                    file.resolve()
                    for file in ds_obj.files
                    if isinstance(file, LocalFile)
                ]
                # Include files in the dictionary
                [
                    required_files[str(file.parent)].add(file.name)
                    for file in files
                ]

    # Convert sets to sorted list to make it more readable
    required_files = {
        key: sorted(values)
        for key, values in required_files.items()
    }
    if not required_files:
        logger.info("No files needed...")
        return

    rel_esarchive = "/gpfs/archive/bsc32/"

    ref_path = next(iter(required_files.keys()))

    # hub will be true when esarchive is mounted in / (ES machines)
    # false if esarchive is mounted in /gpfs/archive/bsc32/ (Nord)
    hub = not ref_path.startswith(rel_esarchive)

    rel_esarchive = "/gpfs/archive/bsc32/"
    # define target folder with environment variables for target folder
    # and user folder
    tfolder = Path(tfolder)

    mk_cmd = []
    cp_cmd = []
    for path, files in required_files.items():
        if hub:
            # relative to '/'
            path_es = path[1:]
            # mounted in gpfs
            path = rel_esarchive + path_es
        else:
            # path already mounted in gpfs
            # relative to rel_esarchive
            path_es = path[len(rel_esarchive):]
        newdir = "$TFOLDER/" + path_es
        # command to make new directories
        mk_cmd.append(f"mkdir -p {newdir}")
        # files to copy
        files = [f"{path}/{file}" for file in files]
        # command to copy files
        cp_cmd.append(" ".join(['dtcp', *files, str(newdir)]))

    # define commands to run
    run_cmd = '\n'.join([f"TFOLDER={tfolder}"]+mk_cmd+cp_cmd)
    # define the file to write the commands in
    cmd_file = Path(Path(recipe_file).stem + "_copy2gpfs.sh")
    cmd_file = cmd_file.resolve()
    # Write the file
    with cmd_file.open("w") as file:
        file.write(run_cmd)
    cmd_file.chmod(0o774)
    # print command to run the file in transfer machine
    if hub:
        run_file = rel_esarchive + str(cmd_file)[1:]
    else:
        run_file = cmd_file
    logger.info("Run the following command in a transfer machine:\nbash %s",
                run_file)


if __name__ == '__main__':
    copy2gpfs(sys.argv[1])
